package cz.fi.muni.pa165.dao;

import cz.fi.muni.pa165.ApplicationContext;
import cz.fi.muni.pa165.entity.Activity;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.annotation.Rollback;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.springframework.transaction.annotation.Transactional;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.PersistenceUnit;
import javax.validation.ConstraintViolationException;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;

import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertNotNull;

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ContextConfiguration(classes = ApplicationContext.class)
@ExtendWith(SpringExtension.class)
class ActivityDaoTest {

    @PersistenceUnit
    private EntityManagerFactory emf;

    @Autowired
    private ActivityDao activityDao;

    private Activity storedActivity;
    private Activity createdActivity;
    private Activity nullNameActivity;
    private Activity nullDateActivity;

    @BeforeAll
    void setUp() {
        storedActivity = new Activity();
        storedActivity.setTime(LocalDateTime.of(2022,3,27,0,0));
        storedActivity.setName("Activity");
        EntityManager em = null;
        try {
            em = emf.createEntityManager();
            em.getTransaction().begin();

            em.persist(storedActivity);

            em.getTransaction().commit();
        } finally {
            if (em != null) {
                em.close();
            }
        }
    }

    @Test
    @Transactional
    @Rollback
    void testCreateActivity() {
        createdActivity = new Activity();
        createdActivity.setTime(LocalDateTime.of(2022,3,26,0,0));
        createdActivity.setName("Creaed activity");
        activityDao.create(createdActivity);
        assertNotNull(activityDao.findById(createdActivity.getId()));
    }

    @Test
    @Transactional
    @Rollback
    void testCreateActivityWithNullName() {
        nullNameActivity = new Activity();
        nullNameActivity.setTime(LocalDateTime.of(2022,3,26,0,0));
        nullNameActivity.setName(null);
        assertThrows(ConstraintViolationException.class, () -> activityDao.create(nullNameActivity));
    }

    @Test
    @Transactional
    @Rollback
    void testCreateActivityWithNullTime() {
        nullDateActivity = new Activity();
        nullDateActivity.setTime(null);
        nullDateActivity.setName("Null date activity");
        assertThrows(ConstraintViolationException.class, () -> activityDao.create(nullDateActivity));
    }

    @Test
    void testFindByID(){
        Activity activity = activityDao.findById(storedActivity.getId());
        assertEquals(storedActivity, activity);
    }

    @Test
    void testFindByNullID(){
        assertThrows(IllegalArgumentException.class, () -> activityDao.findById(null));
    }

    @Test
    void testFindByNonExistingID(){
        Activity activity = activityDao.findById(Long.MAX_VALUE);
        assertNull(activity);
    }

    @Test
    void testFindByName() {
        List<Activity> activities = activityDao.findByName(storedActivity.getName());
        List<Activity> expectedActivities = new ArrayList<>(List.of(storedActivity));
        assertEquals(expectedActivities, activities);
    }

    @Test
    void testFindByNullName() {
        List<Activity> activities = activityDao.findByName(null);
        assertEquals(new ArrayList<>(), activities);
    }


    @Test
    @Transactional
    @Rollback
    void testUpdateActivityName() {
        Activity activity = activityDao.findById(storedActivity.getId());
        activity.setName("New Activity");
        activityDao.update(activity);
        Activity activityWithNewName = activityDao.findById(storedActivity.getId());
        assertEquals("New Activity", activityWithNewName.getName());
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateActivityWithNullName() {
        Activity activity = activityDao.findById(storedActivity.getId());
        activity.setName(null);
        activityDao.update(activity);
        Activity activityWithNullName = activityDao.findById(storedActivity.getId());
        assertNull(activityWithNullName.getName());
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateActivityWithEmptyName() {
        Activity activity = activityDao.findById(storedActivity.getId());
        activity.setName("");
        activityDao.update(activity);
        Activity activityWithEmptyName = activityDao.findById(storedActivity.getId());
        assertEquals("", activityWithEmptyName.getName());
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateActivityTime() {
        Activity activity = activityDao.findById(storedActivity.getId());
        LocalDateTime newDate = LocalDateTime.of(2022,1,1,1,1);
        activity.setTime(newDate);
        activityDao.update(activity);
        Activity activityWithNewTime = activityDao.findById(storedActivity.getId());
        assertEquals(newDate, activityWithNewTime.getTime());
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateActivityWithNullTime() {
        Activity activity = activityDao.findById(storedActivity.getId());
        activity.setTime(null);
        activityDao.update(activity);
        Activity activityWithNullTime = activityDao.findById(storedActivity.getId());
        assertNull(activityWithNullTime.getTime());
    }


    @Test
    @Transactional
    @Rollback
    void testDeleteActivity() {
        activityDao.delete(storedActivity);
        assertNull(activityDao.findById(storedActivity.getId()));
    }

    @Test
    @Transactional
    @Rollback
    void testDeleteNullActivity() {
        assertThrows(NullPointerException.class, () -> activityDao.delete(null));
    }

    @Test
    @Transactional
    @Rollback
    void testDeleteActivityWithoutID() {
        assertThrows(IllegalArgumentException.class, () -> activityDao.delete(new Activity()));
    }


}