package cz.fi.muni.pa165.dao;

import cz.fi.muni.pa165.ApplicationContext;
import cz.fi.muni.pa165.entity.*;
import cz.fi.muni.pa165.enums.BandType;
import cz.fi.muni.pa165.enums.MusicianType;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.annotation.Rollback;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.PersistenceUnit;
import javax.transaction.Transactional;
import javax.validation.ConstraintViolationException;
import java.sql.Time;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;

import static org.junit.jupiter.api.Assertions.*;

/**
 * @author Dávid Želník - 469323
 */
@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ContextConfiguration(classes = ApplicationContext.class)
@ExtendWith(SpringExtension.class)
public class AlbumDaoTest {

    @PersistenceUnit
    private EntityManagerFactory emf;

    @Autowired
    private AlbumDao albumDao;

    private final Album albumOne = new Album();
    private final Album albumTwo = new Album();
    private final Song song = new Song();
    private final Band band = new Band();
    private final Musician musician = new Musician();

    @BeforeAll
    void setUp(){
        song.setSongName("Song name");
        song.setDuration(new Time(22));

        musician.setName("Musician");
        musician.setType(MusicianType.GUITARIST);

        band.setName("Band");
        band.setType(BandType.ROCK);
        band.setBandMembers(new HashSet<>(List.of(musician)));
        band.setBandSongs(new HashSet<>(List.of(song)));

        song.setBand(band);
        musician.setBand(band);

        albumOne.setName("First");
        albumOne.setBand(band);
        albumOne.setSongs(new HashSet<>(List.of(song)));

        albumTwo.setName("Second");
        albumTwo.setBand(band);
        albumTwo.setSongs(new HashSet<>());

        band.setBandAlbums(new HashSet<>(List.of(albumOne, albumTwo)));

        EntityManager em = null;
        try {
            em = emf.createEntityManager();
            em.getTransaction().begin();

            em.persist(musician);
            em.persist(band);
            em.persist(song);

            em.persist(albumOne);
            em.persist(albumTwo);

            em.getTransaction().commit();
        }
        finally {
            if(em != null){
                em.close();
            }
        }
    }

    @Test
    @Transactional
    @Rollback
    void testCreateAlbum() {
        Album newAlbum = new Album();
        newAlbum.setName("Album");
        newAlbum.setBand(band);
        albumDao.create(newAlbum);
        assertNotNull(albumDao.findById(newAlbum.getId()));
    }

    @Test
    @Transactional
    @Rollback
    void testCreateAlbumWithoutName() {
        Album createdAlbum = new Album();
        createdAlbum.setName(null);
        createdAlbum.setBand(band);
        createdAlbum.setSongs(new HashSet<>(List.of(song)));
        assertThrows(ConstraintViolationException.class, () -> albumDao.create(createdAlbum));
    }

    @Test
    @Transactional
    @Rollback
    void testCreateAlbumWithNullBand() {
        Album createdAlbum = new Album();
        createdAlbum.setName("Album");
        createdAlbum.setBand(null);
        createdAlbum.setSongs(new HashSet<>(List.of(song)));
        assertThrows(ConstraintViolationException.class, () -> albumDao.create(createdAlbum));
    }

    @Test
    @Transactional
    @Rollback
    void testCreateEmptyAlbum() {
        Album album = new Album();
        assertThrows(ConstraintViolationException.class, () -> albumDao.create(album));
    }

    @Test
    @Transactional
    @Rollback
    void testFindAlbumById(){
        Album first = albumDao.findById(albumOne.getId());
        Album second = albumDao.findById(albumTwo.getId());
        assertEquals(second, albumTwo);
        assertEquals(first, albumOne);
    }

    @Test
    @Transactional
    @Rollback
    void testFindAlbumByNullId(){
        assertThrows(IllegalArgumentException.class, () -> albumDao.findById(null));
    }

    @Test
    @Transactional
    @Rollback
    void testFindAlbumByNonExistingID(){
        assertNull(albumDao.findById(-4555L));
    }

    @Test
    @Transactional
    @Rollback
    void testFindAlbumByName() {
        List<Album> albumsDb = albumDao.findByName(albumOne.getName());
        List<Album> expectedAlbums = new ArrayList<>(Arrays.asList(albumOne));
        assertEquals(expectedAlbums, albumsDb);
    }

    @Test
    @Transactional
    @Rollback
    void testFindAlbumByNullName() {
        List<Album> albumsDb = albumDao.findByName(null);
        assertEquals(new ArrayList<>(), albumsDb);
    }

    @Test
    @Transactional
    @Rollback
    void testFindAlbumByEmptyName() {
        List<Album> albumsDb = albumDao.findByName("");
        assertEquals(new ArrayList<>(), albumsDb);
    }

    @Test
    @Transactional
    @Rollback
    void testFindNonExistingAlbumByName() {
        List<Album> albumsDb = albumDao.findByName("213456");
        assertEquals(new ArrayList<>(), albumsDb);
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateAlbum() {
        Album albumFromDb = albumDao.findById(albumOne.getId());
        albumFromDb.setName("New album");
        albumDao.update(albumFromDb);
        Album newAlbum = albumDao.findById(albumOne.getId());
        assertEquals("New album", newAlbum.getName());
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateAlbumWithNullName() {
        Album albumInDb = albumDao.findById(albumOne.getId());
        albumInDb.setName(null);
        albumDao.update(albumInDb);
        Album albumWithNullName = albumDao.findById(albumOne.getId());
        assertNull(albumWithNullName.getName());
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateAlbumWithNullBandName() {
        Album albumInDb = albumDao.findById(albumOne.getId());
        albumInDb.setBand(null);
        albumDao.update(albumInDb);
        Album albumWithNullBand = albumDao.findById(albumOne.getId());
        assertNull(albumWithNullBand.getBand());
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateAlbumSongs() {
        Album albumInDb = albumDao.findById(albumOne.getId());
        albumInDb.setSongs(new HashSet<>());
        albumDao.update(albumInDb);
        Album albumWithZeroSongs = albumDao.findById(albumOne.getId());
        assertEquals(new HashSet<>(), albumWithZeroSongs.getSongs());
    }

    @Test
    @Transactional
    @Rollback
    void testDeleteAlbum() {
        albumDao.delete(albumOne);
        assertNull(albumDao.findById(albumOne.getId()));
    }

    @Test
    @Transactional
    @Rollback
    void testDeleteNullAlbum() {
        assertThrows(NullPointerException.class, () -> albumDao.delete(null));
    }

    @Test
    @Transactional
    @Rollback
    void testDeleteAlbumNotInDb() {
        assertThrows(IllegalArgumentException.class, () -> albumDao.delete(new Album()));
    }

}