package cz.fi.muni.pa165.dao;

import cz.fi.muni.pa165.ApplicationContext;
import cz.fi.muni.pa165.entity.Band;
import cz.fi.muni.pa165.entity.Musician;
import cz.fi.muni.pa165.enums.BandType;
import cz.fi.muni.pa165.enums.MusicianType;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.extension.ExtendWith;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.annotation.Rollback;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.springframework.transaction.annotation.Transactional;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.PersistenceUnit;
import javax.validation.ConstraintViolationException;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertNotNull;

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ContextConfiguration(classes = ApplicationContext.class)
@ExtendWith(SpringExtension.class)
class BandDaoTest {

    @PersistenceUnit
    private EntityManagerFactory emf;

    @Autowired
    private BandDao bandDao;

    private Band storedBand;
    private Band createdBand;
    private Band nullNameBand;
    private Band nullTypeBand;
    private Band nullMembersBand;
    private final Set<Musician> bandMembers = new HashSet<>();


    @BeforeAll
    void setUp() {
        storedBand = new Band();
        Musician musicianJohn = new Musician();
        Musician musicianJane = new Musician();
        musicianJane.setName("Jane Smith");
        musicianJane.setType(MusicianType.SINGER);
        musicianJohn.setName("John Doe");
        musicianJohn.setType(MusicianType.GUITARIST);
        bandMembers.add(musicianJane);
        bandMembers.add(musicianJohn);
        storedBand.setName("Band");
        storedBand.setType(BandType.ROCK);
        storedBand.setLogo("New Logo");
        storedBand.setBandMembers(bandMembers);
        EntityManager em = null;
        try {
            em = emf.createEntityManager();
            em.getTransaction().begin();

            em.persist(musicianJane);
            em.persist(musicianJohn);
            em.persist(storedBand);

            em.getTransaction().commit();
        } finally {
            if (em != null) {
                em.close();
            }
        }
    }

    @Test
    void testFindByID(){
        Band bandFromDB = bandDao.findById(storedBand.getId());
        assertEquals(storedBand, bandFromDB);
    }

    @Test
    void testFindByNullID(){
        assertThrows(IllegalArgumentException.class, () -> bandDao.findById(null));
    }

    @Test
    void testFindByNonExistingID(){
        Band bandFromDB = bandDao.findById(Long.MAX_VALUE);
        assertNull(bandFromDB);
    }

    @Test
    void testFindByName() {
        List<Band> bandsFromDB = bandDao.findByName(storedBand.getName());
        List<Band> expectedBands = new ArrayList<>(List.of(storedBand));
        assertEquals(expectedBands, bandsFromDB);
    }

    @Test
    void testFindByNullName() {
        List<Band> bandsFromDB = bandDao.findByName(null);
        assertEquals(new ArrayList<>(), bandsFromDB);
    }

    @Test
    @Transactional
    @Rollback
    void testCreateBand() {
        createdBand = new Band();
        createdBand.setBandMembers(bandMembers);
        createdBand.setName("created band");
        createdBand.setType(BandType.ROCK);
        bandDao.create(createdBand);
        assertNotNull(bandDao.findById(createdBand.getId()));
    }

    @Test
    @Transactional
    @Rollback
    void testCreateBandWithNullName() {
        nullNameBand = new Band();
        nullNameBand.setBandMembers(bandMembers);
        nullNameBand.setName(null);
        nullNameBand.setType(BandType.ROCK);
        assertThrows(ConstraintViolationException.class, () -> bandDao.create(nullNameBand));
    }

    @Test
    @Transactional
    @Rollback
    void testCreateBandWithNullType() {
        nullTypeBand = new Band();
        nullTypeBand.setBandMembers(bandMembers);
        nullTypeBand.setName("Null type band");
        nullTypeBand.setType(null);
        assertThrows(ConstraintViolationException.class, () -> bandDao.create(nullTypeBand));
    }

    @Test
    @Transactional
    @Rollback
    void testCreateBandWithNullMembers() {
        nullMembersBand = new Band();
        nullMembersBand.setBandMembers(null);
        nullMembersBand.setName("Null members band");
        nullMembersBand.setType(BandType.ROCK);
        assertThrows(ConstraintViolationException.class, () -> bandDao.create(nullMembersBand));
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateBandName() {
        Band bandFromDB = bandDao.findById(storedBand.getId());
        bandFromDB.setName("New Band");
        bandDao.update(bandFromDB);
        Band bandWithNewName = bandDao.findById(storedBand.getId());
        assertEquals("New Band", bandWithNewName.getName());
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateBandWithNullName() {
        Band bandFromDB = bandDao.findById(storedBand.getId());
        bandFromDB.setName(null);
        bandDao.update(bandFromDB);
        Band bandWithNullName = bandDao.findById(storedBand.getId());
        assertNull(bandWithNullName.getName());
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateBandWithEmptyName() {
        Band bandFromDB = bandDao.findById(storedBand.getId());
        bandFromDB.setName("");
        bandDao.update(bandFromDB);
        Band bandWithNullName = bandDao.findById(storedBand.getId());
        assertEquals("", bandWithNullName.getName());
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateBandType() {
        Band bandFromDB = bandDao.findById(storedBand.getId());
        bandFromDB.setType(BandType.COUNTRY);
        bandDao.update(bandFromDB);
        Band bandWithNewType = bandDao.findById(storedBand.getId());
        assertEquals(BandType.COUNTRY, bandWithNewType.getType());
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateBandWithNullType() {
        Band bandFromDB = bandDao.findById(storedBand.getId());
        bandFromDB.setType(null);
        bandDao.update(bandFromDB);
        Band bandWithNullType = bandDao.findById(storedBand.getId());
        assertNull(bandWithNullType.getType());
    }

    @Test
    @Transactional
    @Rollback
    void testDeleteBand() {
        bandDao.delete(storedBand);
        assertNull(bandDao.findById(storedBand.getId()));
    }

    @Test
    @Transactional
    @Rollback
    void testDeleteNullBand() {
        assertThrows(NullPointerException.class, () -> bandDao.delete(null));
    }

    @Test
    @Transactional
    @Rollback
    void testDeleteBandWithoutID() {
        assertThrows(IllegalArgumentException.class, () -> bandDao.delete(new Band()));
    }


}