package cz.fi.muni.pa165.dao;


import cz.fi.muni.pa165.ApplicationContext;
import cz.fi.muni.pa165.entity.Album;
import cz.fi.muni.pa165.entity.Band;
import cz.fi.muni.pa165.entity.Manager;
import cz.fi.muni.pa165.entity.Musician;
import cz.fi.muni.pa165.enums.BandType;
import cz.fi.muni.pa165.enums.MusicianType;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.annotation.Rollback;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.PersistenceUnit;
import javax.transaction.Transactional;
import javax.validation.ConstraintViolationException;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import static org.junit.jupiter.api.Assertions.*;

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ContextConfiguration(classes = ApplicationContext.class)
@ExtendWith(SpringExtension.class)
public class ManagerDaoTest {

    @PersistenceUnit
    private EntityManagerFactory emf;

    @Autowired
    private ManagerDao managerDao;

    private final Manager manager = new Manager();
    private final Manager manager2 = new Manager();
    private final Band band = new Band();
    private final Set<Musician> bandMembers = new HashSet<>();
    private final Band band2 = new Band();


    @BeforeAll
    void setUp() {
        // create first band
        manager.setName("Jef the manager");
        band.setName("Band 1");
        band.setType(BandType.ROCK);
        band.setLogo("New Logo");
        band.setBandMembers(bandMembers);
        manager.setManagedBands(new HashSet<>(List.of(band)));

        //create second band
        manager2.setName("Logan the manager");
        band2.setName("Original Name");
        band2.setType(BandType.ROCK);
        band2.setLogo("Original Logo");
        band2.setBandMembers(bandMembers);
        manager2.setManagedBands(new HashSet<>(List.of(band2)));


        EntityManager em = null;
        try {
            //create entity manager
            em = emf.createEntityManager();
            em.getTransaction().begin();

            //add to table
            em.persist(band);
            em.persist(band2);
            em.persist(manager);
            em.persist(manager2);

            //commit added data
            em.getTransaction().commit();
        } finally {
            //after commit close manager
            if (em != null) {
                em.close();
            }
        }
    }

    @Test
    void testFindByID(){
        Manager managerFromDB = managerDao.findById(manager.getId());
        Manager manager2FromDB = managerDao.findById(manager2.getId());
        assertEquals(manager.getName(), managerFromDB.getName());
        assertEquals(manager2.getName(), manager2FromDB.getName());
    }

    @Test
    void testFindByNullID(){
        assertThrows(IllegalArgumentException.class, () -> managerDao.findById(null));
    }

    @Test
    void testFindByNonExistingID(){
        Manager manager = managerDao.findById(-1L);
        assertNull(manager);
    }

    @Test
    void testFindByName() {
        List<Manager> managerFromDB = managerDao.findByName(manager.getName());
        List<Manager> expectedManagers = new ArrayList<>(List.of(manager));
        assertEquals(expectedManagers, managerFromDB);
    }

    @Test
    void testFindByNullName() {
        List<Manager> managerFromDB = managerDao.findByName(null);
        assertEquals(new ArrayList<>(), managerFromDB);
    }

    @Test
    void testFindByEmptyName() {
        List<Manager> managerFromDB = managerDao.findByName("");
        assertEquals(new ArrayList<>(), managerFromDB);
    }

    @Test
    @Transactional
    @Rollback
    void testCreateManager() {
        Manager testManager = new Manager();
        testManager.setName("Test Band");
        Band testBand = new Band();
        testManager.setManagedBands(new HashSet<>(List.of(testBand)));
        managerDao.create(testManager);
        assertNotNull(managerDao.findById(testManager.getId()));
    }


    @Test
    @Transactional
    @Rollback
    void testCreateManagerWithoutName() {
        Manager testManager = new Manager();
        Band testBand = new Band();
        testManager.setName(null);
        testManager.setManagedBands(new HashSet<>(List.of(testBand)));
        testBand.setType(BandType.METAL);
        assertThrows(ConstraintViolationException.class, () -> managerDao.create(testManager));
    }

    @Test
    @Transactional
    @Rollback
    void testCreateManagerWithoutBand() {
        Manager testManager = new Manager();
        testManager.setName("Name of manager");
        managerDao.create(testManager);
        assertNotNull(managerDao.findById(testManager.getId()));
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateManager() {
        Manager managerFromDB = managerDao.findById(manager.getId());
        managerFromDB.setName("New Manager");
        managerDao.update(managerFromDB);
        Manager newManager = managerDao.findById(band.getId());
        assertEquals("New Manager", newManager.getName());
    }


    //Failing maybe check if the name is not null before update
    @Test
    @Transactional
    @Rollback
    void testUpdateManagerWithNullName() {
        Manager managerFromDB = managerDao.findById(manager.getId());
        managerFromDB.setName(null);
        managerDao.update(managerFromDB);
        Manager managerWithNullName = managerDao.findById(manager.getId());
        assertNull(managerWithNullName.getName());
    }

    @Test
    @Transactional
    @Rollback
    void testDeleteManager() {
        managerDao.delete(manager);
        assertNull(managerDao.findById(manager.getId()));
    }
    @Test
    @Transactional
    @Rollback
    void testDeleteNullManager() {
        assertThrows(NullPointerException.class, () ->managerDao.delete(null));
    }

    @Test
    @Transactional
    @Rollback
    void testDeleteManagerWithoutID() {
        assertThrows(IllegalArgumentException.class, () -> managerDao.delete(new Manager()));
    }
}
