package cz.fi.muni.pa165.dao;

import cz.fi.muni.pa165.ApplicationContext;
import cz.fi.muni.pa165.entity.*;
import cz.fi.muni.pa165.enums.MusicianType;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.annotation.Rollback;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.PersistenceUnit;
import javax.transaction.Transactional;
import javax.validation.ConstraintViolationException;
import java.util.ArrayList;
import java.util.List;

import static org.junit.jupiter.api.Assertions.*;

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ContextConfiguration(classes = ApplicationContext.class)
@ExtendWith(SpringExtension.class)
public class MusicianDaoTest {
    @PersistenceUnit
    private EntityManagerFactory emf;

    @Autowired
    private MusicianDao musicianDao;

    private final Musician musicianStephen = new Musician();
    private final Musician musicianMick = new Musician();

    @BeforeAll
    void setUp(){
        musicianStephen.setName("Stephen");
        musicianStephen.setType(MusicianType.GUITARIST);

        musicianMick.setName("Mick");
        musicianMick.setType(MusicianType.SINGER);

        EntityManager em = null;
        try {
            em = emf.createEntityManager();
            em.getTransaction().begin();
            em.persist(musicianStephen);
            em.persist(musicianMick);

            em.getTransaction().commit();
        }
        finally {
            if(em != null){
                em.close();
            }
        }
    }

    @Test
    void testFindMusicianById(){
        Musician musicianStephenFromDb = musicianDao.findById(musicianStephen.getId());
        Musician musicianMickFromDb = musicianDao.findById(musicianMick.getId());
        assertEquals(musicianStephenFromDb, musicianStephen);
        assertEquals(musicianMickFromDb, musicianMick);
    }

    @Test
    void findMusicianByNullId(){
        assertThrows(IllegalArgumentException.class, () -> musicianDao.findById(null));
    }

    @Test
    void testFindMusicianByNonExistingID(){
        assertNull(musicianDao.findById(-1L));
    }

    @Test
    void testFindMusicianByName() {
        List<Musician> musicianFromDb = musicianDao.findByName(musicianStephen.getName());
        List<Musician> expectedMusicians = new ArrayList<>(List.of(musicianStephen));
        assertEquals(expectedMusicians, musicianFromDb);
    }

    @Test
    void testFindMusicianByNullName() {
        List<Musician> musicianFromDB = musicianDao.findByName(null);
        assertEquals(new ArrayList<>(), musicianFromDB);
    }

    @Test
    void testFindMusicianByEmptyName() {
        List<Musician> musicianFromDB = musicianDao.findByName("");
        assertEquals(new ArrayList<>(), musicianFromDB);
    }

    @Test
    @Transactional
    @Rollback
    void testCreateMusician() {
        Musician musician = new Musician();
        musician.setName("Test");
        musician.setType(MusicianType.SINGER);
        musicianDao.create(musician);
        assertNotNull(musicianDao.findById(musician.getId()));
    }

    @Test
    @Transactional
    @Rollback
    void testCreateMusicianWithoutName() {
        Musician musician = new Musician();
        musician.setName(null);
        assertThrows(ConstraintViolationException.class, () -> musicianDao.create(musician));
    }

    @Test
    @Transactional
    @Rollback
    void testCreateMusicianWithoutType() {
        Musician musician = new Musician();
        musician.setType(null);
        assertThrows(ConstraintViolationException.class, () -> musicianDao.create(musician));
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateMusician() {
        Musician musicianFromDb = musicianDao.findById(musicianStephen.getId());
        musicianFromDb.setName("Stephanos");
        musicianFromDb.setType(MusicianType.DRUMMER);
        musicianDao.update(musicianFromDb);
        Musician newMusician = musicianDao.findById(musicianStephen.getId());
        assertEquals("Stephanos", newMusician.getName());
        assertEquals(MusicianType.DRUMMER, newMusician.getType());
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateMusicianWithNullName() {
        Musician musicianFromDb = musicianDao.findById(musicianStephen.getId());
        musicianFromDb.setName(null);
        musicianDao.update(musicianFromDb);
        Musician musicianWithNullName = musicianDao.findById(musicianStephen.getId());
        assertNull(musicianWithNullName.getName());
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateMusicianWithNullType() {
        Musician musicianFromDb = musicianDao.findById(musicianStephen.getId());
        musicianFromDb.setType(null);
        musicianDao.update(musicianFromDb);
        Musician musicianWithNullType = musicianDao.findById(musicianStephen.getId());
        assertNull(musicianWithNullType.getType());
    }

    @Test
    @Transactional
    @Rollback
    void testDeleteMusician() {
        musicianDao.delete(musicianStephen);
        assertNull(musicianDao.findById(musicianStephen.getId()));
    }

    @Test
    @Transactional
    @Rollback
    void testDeleteNullMusician() {
        assertThrows(NullPointerException.class, () -> musicianDao.delete(null));
    }

    @Test
    @Transactional
    @Rollback
    void testDeleteMusicianWithoutID() {
        assertThrows(IllegalArgumentException.class, () -> musicianDao.delete(new Musician()));
    }
}
