package cz.fi.muni.pa165.dao;

import cz.fi.muni.pa165.ApplicationContext;
import cz.fi.muni.pa165.entity.*;
import cz.fi.muni.pa165.enums.BandType;
import cz.fi.muni.pa165.enums.MusicianType;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.annotation.Rollback;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.PersistenceUnit;
import javax.transaction.Transactional;
import javax.validation.ConstraintViolationException;
import java.sql.Time;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;

import static org.junit.jupiter.api.Assertions.*;

/**
 * @author Dávid Želník - 469323
 */
@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@ContextConfiguration(classes = ApplicationContext.class)
@ExtendWith(SpringExtension.class)
public class SongDaoTest {

    @PersistenceUnit
    private EntityManagerFactory emf;

    @Autowired
    private SongDao songDao;

    private final Song songOne = new Song();
    private final Song songTwo = new Song();
    private final Band band = new Band();
    private final Musician musician = new Musician();

    @BeforeAll
    void setUp(){
        musician.setName("Musician");
        musician.setType(MusicianType.GUITARIST);

        songOne.setSongName("Song One");
        songOne.setDuration(new Time(22));
        songTwo.setSongName("Song Two");
        songTwo.setDuration(new Time(33));

        band.setName("Band");
        band.setType(BandType.ROCK);
        band.setBandMembers(new HashSet<>(List.of(musician)));
        band.setBandSongs(new HashSet<>(List.of(songOne, songTwo)));

        songOne.setBand(band);
        songTwo.setBand(band);

        EntityManager em = null;
        try {
            em = emf.createEntityManager();
            em.getTransaction().begin();


            em.persist(band);
            em.persist(musician);
            em.persist(songTwo);
            em.persist(songOne);

            em.getTransaction().commit();
        }
        finally {
            if(em != null){
                em.close();
            }
        }
    }

    @Test
    @Transactional
    @Rollback
    void testCreateSong() {
        Song newSong = new Song();
        newSong.setSongName("Song");
        newSong.setBand(band);
        newSong.setDuration(new Time(548));
        songDao.create(newSong);
        assertNotNull(songDao.findById(newSong.getId()));
    }

    @Test
    @Transactional
    @Rollback
    void testCreateSongWithNullName() {
        Song createdSong = new Song();
        createdSong.setSongName(null);
        createdSong.setBand(band);
        createdSong.setDuration(new Time(548));
        assertThrows(ConstraintViolationException.class, () -> songDao.create(createdSong));
    }

    @Test
    @Transactional
    @Rollback
    void testCreateSongWithNullBand() {
        Song createdSong = new Song();
        createdSong.setSongName("Songs");
        createdSong.setBand(null);
        createdSong.setDuration(new Time(548));
        assertThrows(ConstraintViolationException.class, () -> songDao.create(createdSong));
    }


    @Test
    @Transactional
    @Rollback
    void testCreateSongWithNullTime() {
        Song createdSong = new Song();
        createdSong.setSongName("Songs");
        createdSong.setBand(band);
        createdSong.setDuration(null);
        assertThrows(ConstraintViolationException.class, () -> songDao.create(createdSong));
    }

    @Test
    @Transactional
    @Rollback
    void testCreateEmptySong() {
        Song song = new Song();
        assertThrows(ConstraintViolationException.class, () -> songDao.create(song));
    }

    @Test
    @Transactional
    @Rollback
    void testFindSongById(){
        Song first = songDao.findById(songOne.getId());
        Song second = songDao.findById(songTwo.getId());
        assertEquals(second, songTwo);
        assertEquals(first, songOne);
    }

    @Test
    @Transactional
    @Rollback
    void testFindSongByNullId(){
        assertThrows(IllegalArgumentException.class, () -> songDao.findById(null));
    }

    @Test
    @Transactional
    @Rollback
    void testFindSongByNonExistingID(){
        assertNull(songDao.findById(-4555L));
    }

    @Test
    @Transactional
    @Rollback
    void testFindSongByName() {
        List<Song> songsDb = songDao.findByName(songOne.getSongName());
        List<Song> expectedSongs = new ArrayList<>(Arrays.asList(songOne));
        assertEquals(expectedSongs, songsDb);
    }

    @Test
    @Transactional
    @Rollback
    void testFindSongByNullName() {
        List<Song> songsDb = songDao.findByName(null);
        assertEquals(new ArrayList<>(), songsDb);
    }

    @Test
    @Transactional
    @Rollback
    void testFindSongByEmptyName() {
        List<Song> songsDb = songDao.findByName("");
        assertEquals(new ArrayList<>(), songsDb);
    }

    @Test
    @Transactional
    @Rollback
    void testFindNonExistingSongByName() {
        List<Song> songsDb = songDao.findByName("213456");
        assertEquals(new ArrayList<>(), songsDb);
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateSong() {
        Song songFromDb = songDao.findById(songOne.getId());
        songFromDb.setSongName("New song");
        songDao.update(songFromDb);
        Song newSong = songDao.findById(songOne.getId());
        assertEquals("New song", newSong.getSongName());
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateSongWithNullName() {
        Song songInDb = songDao.findById(songOne.getId());
        songInDb.setSongName(null);
        songDao.update(songInDb);
        Song songWithNullName = songDao.findById(songOne.getId());
        assertNull(songWithNullName.getSongName());
    }

    @Test
    @Transactional
    @Rollback
    void testUpdateSongWithNullBand() {
        Song songInDb = songDao.findById(songOne.getId());
        songInDb.setBand(null);
        songDao.update(songInDb);
        Song songWithNullBandName = songDao.findById(songOne.getId());
        assertNull(songWithNullBandName.getBand());
    }


    @Test
    @Transactional
    @Rollback
    void testUpdateSongWithNullDuration() {
        Song songInDb = songDao.findById(songOne.getId());
        songInDb.setDuration(null);
        songDao.update(songInDb);
        Song songWithNullDuration = songDao.findById(songOne.getId());
        assertNull(songWithNullDuration.getDuration());
    }

    @Test
    @Transactional
    @Rollback
    void testDeleteSong() {
        songDao.delete(songOne);
        assertNull(songDao.findById(songOne.getId()));
    }

    @Test
    @Transactional
    @Rollback
    void testDeleteNullSong() {
        assertThrows(NullPointerException.class, () -> songDao.delete(null));
    }

    @Test
    @Transactional
    @Rollback
    void testDeleteSongNotInDb() {
        assertThrows(IllegalArgumentException.class, () -> songDao.delete(new Song()));
    }

}